class VerificationError(Exception):
    def __init__(self, failed_components: list[str]):
        self.failed_components = failed_components
        super().__init__(f"Verification failed at: {', '.join(failed_components)}")



from fastecdsa.curve import P384

from pyivxv.encoding.message import encode_to_point, decode_from_point


def test_encoding():
    curve = P384

    message = "0000.101"
    encoded = encode_to_point(message.encode(), curve)
    decoded = decode_from_point(encoded, curve).decode()

    assert decoded == message



from fastecdsa.point import Point
from pyasn1.codec.der import encoder, decoder
from pyasn1_modules import rfc5280

from pyivxv.asn1.schemas import ECCElGamalDecryptionProof, ECCElGamalDecryptionChallenge, ElGamalCiphertextInfo
from pyivxv.crypto import drbg
from pyivxv.crypto.ciphertext import ElGamalCiphertext
from pyivxv.crypto.exceptions import VerificationError
from pyivxv.crypto.keys import PublicKey
from pyivxv.encoding.ec_point import point_to_bytes, point_from_bytes


class DecryptionProof:
    mComm: Point
    kComm: Point
    response: int

    def __init__(self, mc: Point, kc: Point, res: int):
        self.mComm = mc
        self.kComm = kc
        self.response = res

    def to_bytes(self) -> bytes:
        adp = ECCElGamalDecryptionProof()
        adp["aMsgCommitment"] = point_to_bytes(self.mComm)
        adp["bKeyCommitment"] = point_to_bytes(self.kComm)
        adp["sResponse"] = self.response
        return encoder.encode(adp)

    @classmethod
    def from_bytes(cls, data: bytes):
        dp, _ = decoder.decode(data, asn1Spec=ECCElGamalDecryptionProof())
        mComm = point_from_bytes(dp["aMsgCommitment"])
        kComm = point_from_bytes(dp["bKeyCommitment"])
        response = int(dp["sResponse"])
        return cls(mComm, kComm, response)

    def verify(self, M: Point, ct: ElGamalCiphertext, pk: PublicKey) -> None:
        message_bytes = point_to_bytes(M)
        mc_bytes = point_to_bytes(self.mComm)
        kc_bytes = point_to_bytes(self.kComm)

        seed = derive_seed(pk.spki, ct.to_asn1(), message_bytes, mc_bytes, kc_bytes)
        challenge = drbg.randbelow(seed, pk.curve.q)

        failed = []

        lhs1 = self.response * ct.U
        rhs1 = self.mComm + (ct.V - M) * challenge
        if lhs1 != rhs1:
            failed.append("aMsgCommitment")

        lhs2 = self.response * pk.curve.G
        rhs2 = self.kComm + pk.H * challenge
        if lhs2 != rhs2:
            failed.append("bKeyCommitment")

        if failed:
            raise VerificationError(failed)


def derive_seed(pub: rfc5280.SubjectPublicKeyInfo, enc: ElGamalCiphertextInfo, dec: bytes,
                msg_commitment: bytes, key_commitment: bytes) -> bytes:
    seed = ECCElGamalDecryptionChallenge()
    seed["niProofDomain"] = "DECRYPTION"
    seed["publicKey"] = pub
    seed["ciphertextInfo"] = enc
    seed["encodedPlaintext"] = dec
    seed["aMsgCommitment"] = msg_commitment
    seed["bKeyCommitment"] = key_commitment

    return encoder.encode(seed)



from fastecdsa.curve import P384
from fastecdsa.point import Point
from pyasn1.codec.der import decoder, encoder
from pyasn1.type.univ import OctetString


def point_from_bytes(der: bytes | OctetString) -> Point:
    """Recover a curve point from its SEC 1 uncompressed byte representation."""
    if isinstance(der, OctetString):
        der = der.asOctets()

    # The first byte indicates whether the point is compressed.
    # In IVXV, no compression is used (0x04).
    concatenated = der[1:]
    half_len = len(concatenated) // 2
    assert half_len == 48, "Public key is not P-384"
    x = int.from_bytes(concatenated[:half_len], "big")
    y = int.from_bytes(concatenated[half_len:], "big")
    return Point(x, y, curve=P384)


def point_to_bytes(P: Point) -> bytes:
    """Get the SEC 1 uncompressed byte representation of the curve point."""
    bl = (P.curve.p.bit_length() + 7) // 8
    x_bytes = P.x.to_bytes(bl, "big")
    y_bytes = P.y.to_bytes(bl, "big")
    return b"\x04" + x_bytes + y_bytes


def point_from_der(der: bytes | OctetString) -> Point:
    """Extract a curve point from its DER-encoded SEC 1 byte representation.

    Use this method only if the point bytes are prefixed with the
    OCTET STRING type. Otherwise, use `point_from_bytes()`.
    """
    if isinstance(der, OctetString):
        der = der.asOctets()

    point_bytes, _ = decoder.decode(der, asn1Spec=OctetString())
    return point_from_bytes(point_bytes)


def point_to_der(point: Point) -> bytes:
    """Get the DER-encoded SEC 1 uncompressed byte representation of the curve point.

    Use this method only if you need the point bytes as a valid ASN1 DER object,
    i.e. prefixed with the OCTET STRING type. Otherwise, use `point_to_bytes()`.
    """
    ecp = OctetString(point_to_bytes(point))
    return encoder.encode(ecp)



from fastecdsa.curve import Curve
from fastecdsa.point import Point

from pyivxv.math.ntheory import legendre, tonelli

PADDING_HEAD = [
    [0x00],
    [0b00111111],
    [0b00011111],
    [0b00001111],
    [0b00000111],
    [0b00000011],
    [0b00000001],
    [0b00000000, 0b11111111],
    [0b00000000, 0b01111111],
    [0b00000000, 0b00111111],
    [0b00000000, 0b00011111]
]
PADDING_FILL = b"\xFF"
PADDING_END = b"\xFE"
IVXV_SHIFT = 10


def encode_to_point(message: bytes, curve: Curve, shift=IVXV_SHIFT) -> Point:
    curve_byte_len = (curve.p.bit_length() + 7) // 8

    padding_head = PADDING_HEAD[shift]
    padding_len = curve_byte_len - len(padding_head) - len(message) - 1
    padded = bytes(padding_head) + (PADDING_FILL * padding_len) + PADDING_END + message

    candidate = int.from_bytes(padded, "big") << shift
    max_tries = 2 ** shift
    for _ in range(max_tries):
        # y^2 = x^3 + ax + b
        rhs = (pow(candidate, 3, curve.p) + curve.a * candidate + curve.b) % curve.p
        if legendre(rhs, curve.p) == 1:
            y = tonelli(rhs, curve.p)
            return Point(candidate, y, curve)
        candidate += 1

    raise RuntimeError("could not encode the data as a curve point")


def decode_from_point(M: Point, curve: Curve, shift=IVXV_SHIFT) -> bytes:
    curve_byte_len = (curve.p.bit_length() + 7) // 8

    x_bytes = M.x.to_bytes(curve_byte_len, "big")

    # The padded message is structured as follows: 01 head || padding 0 || message.
    # The leading (padding) bits must be 0b01.
    # The padding head is necessary for byte-length alignment and length. It consists of 1-bits that follow
    # the leading 0b01 and 0-bits that precede the leading 0b01.
    # The padding itself consists of 1-bits, and is terminated by a 0-bit ([0xFF]* 0xFE after right-shifting).
    if (x_bytes[0] & 0b11000000) != 0b01000000:
        raise RuntimeError("incorrect leading plaintext padding bits")

    tmp = M.x >> shift
    padded = tmp.to_bytes(curve_byte_len, "big")

    # Verify the padding header to ensure that the shift is correct.
    head = PADDING_HEAD[shift]
    if bytes(head) != padded[:len(head)]:
        raise RuntimeError("incorrect padding heading")

    try:
        # 0xFE = 0b11111110 as the byte-aligned padding is terminated by a 0-bit.
        padding_end = padded.index(0xfe)
    except ValueError:
        raise RuntimeError("padding end not found")

    unpadded = padded[padding_end + 1:]
    return unpadded



import base64
import re


def pem_to_der(data: bytes) -> bytes:
    if b"-----BEGIN" not in data:
        return data

    pem_header_re = re.compile(br"-----BEGIN ([A-Z0-9 ]+)-----")
    pem_footer_re = re.compile(br"-----END ([A-Z0-9 ]+)-----")

    header_match = pem_header_re.search(data)
    footer_match = pem_footer_re.search(data)

    if not (header_match and footer_match):
        raise ValueError("Invalid PEM: missing header or footer")

    header_type = header_match.group(1)
    footer_type = footer_match.group(1)

    if header_type != footer_type:
        raise ValueError(
            f"PEM header/footer mismatch: {header_type.decode()} vs {footer_type.decode()}"
        )

    body = data[header_match.end():footer_match.start()]
    body = b"".join(body.split())

    try:
        return base64.b64decode(body, validate=True)
    except Exception as e:
        raise ValueError("Invalid base64 in PEM body") from e



import pytest

from pyivxv.crypto.exceptions import VerificationError
from pyivxv.crypto.keys import generate_private_key
from pyivxv.crypto.zkp import generate_decryption_proof
from pyivxv.encoding.message import encode_to_point, decode_from_point


def test_unblinding():
    sk = generate_private_key()
    pk = sk.public_key

    message = "0000.101"

    M = encode_to_point(message.encode(), pk.curve)
    ct = pk.encrypt(M, store_ephemeral=True)
    dec = ct.unblind(pk.H)

    decoded = decode_from_point(dec, pk.curve).decode()

    assert dec == M
    assert decoded == message


def test_decryption():
    sk = generate_private_key()
    pk = sk.public_key

    message = "0000.101"

    ct = pk.encode_and_encrypt(message)
    pt = sk.decrypt_and_decode(ct)

    assert pt == message


def test_provable_decryption():
    sk = generate_private_key()
    pk = sk.public_key

    message = "0000.101"

    ct = pk.encode_and_encrypt(message)
    M = sk.decrypt(ct)

    proof = generate_decryption_proof(M, ct, sk)

    assert proof.verify(M, ct, pk) is None

    with pytest.raises(VerificationError):
        proof.verify(M * 2, ct, pk)



import base64
import secrets
from textwrap import wrap
from typing import Literal

from fastecdsa import keys
from fastecdsa.curve import P384
from fastecdsa.point import Point
from pyasn1.codec.der import decoder as der_decoder, encoder as der_encoder
from pyasn1.type import tag
from pyasn1.type.univ import BitString
from pyasn1_modules import rfc5280, rfc5915, rfc5480

from pyivxv.asn1.schemas import ECCElGamalPublicKey, ECCElGamalParameters, id_ivxv_ecc_elgamal
from pyivxv.crypto.ciphertext import ElGamalCiphertext
from pyivxv.encoding.ec_point import point_from_bytes, point_to_der
from pyivxv.encoding.message import encode_to_point, decode_from_point
from pyivxv.encoding.pem import pem_to_der


class PublicKey:
    def __init__(self, H: Point, election_id: str, spki: rfc5280.SubjectPublicKeyInfo | None = None,
                 lifted: bool = True):
        self.H = H
        self.election_id = election_id
        self.lifted = lifted

        if spki is None:
            spki = self._to_asn1()

        self.spki = spki

    @classmethod
    def from_public_bytes(cls, data: bytes):
        spki, _ = der_decoder.decode(pem_to_der(data), asn1Spec=rfc5280.SubjectPublicKeyInfo())
        pkey, _ = der_decoder.decode(spki["subjectPublicKey"].asOctets(), asn1Spec=ECCElGamalPublicKey())
        params, _ = der_decoder.decode(spki["algorithm"]["parameters"].asOctets(), asn1Spec=ECCElGamalParameters())
        lifted = bool(params["lifted"]) if params["lifted"].hasValue() else False
        return cls(point_from_bytes(pkey["pubY"]), params["electionId"], spki, lifted)

    def public_bytes(self, encoding: Literal["DER", "PEM"] = "DER") -> bytes:
        if encoding != "DER" and encoding != "PEM":
            raise ValueError("Unsupported encoding")

        der = der_encoder.encode(self.spki)
        if encoding == "DER":
            return der

        b64 = base64.b64encode(der).decode("ascii")
        pem_lines = wrap(b64, 64)
        pem_lines = ["-----BEGIN PUBLIC KEY-----"] + pem_lines + ["-----END PUBLIC KEY-----"]
        pem_bytes = "\n".join(pem_lines).encode("ascii")
        return pem_bytes

    def encrypt(self, M: Point, store_ephemeral=False) -> ElGamalCiphertext:
        r = secrets.randbelow(self.curve.q)
        return ElGamalCiphertext(self.curve.G * r, M + (self.H * r),
                                 r if store_ephemeral else None)

    def lifted_encrypt(self, m: int, store_ephemeral=False) -> ElGamalCiphertext:
        lifted_message = self.curve.G * m
        return self.encrypt(lifted_message, store_ephemeral)

    def encode_and_encrypt(self, m: str, store_ephemeral=False) -> ElGamalCiphertext:
        encoded = encode_to_point(m.encode(), self.curve)
        return self.encrypt(encoded, store_ephemeral)

    def _to_asn1(self) -> rfc5280.SubjectPublicKeyInfo:
        params = ECCElGamalParameters()
        params["curve"] = "P-384"
        params["electionId"] = self.election_id
        params["lifted"] = self.lifted

        spki = rfc5280.SubjectPublicKeyInfo()
        spki["algorithm"]["algorithm"] = id_ivxv_ecc_elgamal
        spki["algorithm"]["parameters"] = params

        x_bytes = self.H.x.to_bytes(48, "big")
        y_bytes = self.H.y.to_bytes(48, "big")
        ec_point = b"\x04" + x_bytes + y_bytes

        pk = ECCElGamalPublicKey()
        pk["pubY"] = ec_point
        pk_der = der_encoder.encode(pk)

        spki["subjectPublicKey"] = BitString.fromOctetString(pk_der)
        return spki

    @property
    def curve(self):
        return self.H.curve


class PrivateKey:
    def __init__(self, x: int, election_id: str, curve=P384):
        self.x = x
        self.election_id = election_id
        self.curve = curve
        self.public_key = PublicKey(self.x * self.curve.G, self.election_id)

    @classmethod
    def from_private_bytes(cls, data: bytes, election_id: str, curve=P384):
        ecpk, _ = der_decoder.decode(pem_to_der(data), asn1Spec=rfc5915.ECPrivateKey())
        x = int.from_bytes(ecpk["privateKey"].asOctets(), "big")
        return cls(x, election_id, curve)

    def private_bytes(self, encoding: Literal["DER", "PEM"] = "DER") -> bytes:
        if encoding != "DER" and encoding != "PEM":
            raise ValueError("Unsupported encoding")

        ec_params = rfc5480.ECParameters().subtype(
            explicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
        )
        ec_params["namedCurve"] = rfc5480.secp384r1

        ec_pub = BitString().fromOctetString(point_to_der(self.public_key.H)).subtype(
            explicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1))

        epk = rfc5915.ECPrivateKey()
        epk["version"] = 1
        epk["privateKey"] = self.x.to_bytes(48, "big")
        epk["parameters"] = ec_params
        epk["publicKey"] = ec_pub

        der = der_encoder.encode(epk)
        if encoding == "DER":
            return der

        pem_type = "EC PRIVATE KEY"

        b64 = base64.b64encode(der).decode("ascii")
        pem_lines = wrap(b64, 64)
        pem_lines = [f"-----BEGIN {pem_type}-----"] + pem_lines + [f"-----END {pem_type}-----"]
        pem_bytes = "\n".join(pem_lines).encode("ascii")
        return pem_bytes

    def decrypt(self, ct: ElGamalCiphertext) -> Point:
        D = self.x * ct.U
        return ct.V - D

    def decrypt_and_decode(self, ct: ElGamalCiphertext) -> str:
        M = self.decrypt(ct)
        m_bytes = decode_from_point(M, self.curve)
        return m_bytes.decode()


def generate_private_key(election_id: str = "TEST") -> PrivateKey:
    x, H = keys.gen_keypair(P384)
    return PrivateKey(x, election_id)



import secrets

from fastecdsa.point import Point

from pyivxv.crypto import drbg
from pyivxv.crypto.ciphertext import ElGamalCiphertext
from pyivxv.crypto.decryption_proof import DecryptionProof, derive_seed
from pyivxv.crypto.keys import PrivateKey
from pyivxv.encoding.ec_point import point_to_bytes


def generate_decryption_proof(M: Point, ct: ElGamalCiphertext, sk: PrivateKey) -> DecryptionProof:
    t = secrets.randbelow(sk.curve.q)
    message_commitment = ct.U * t
    key_commitment = sk.curve.G * t

    message_bytes = point_to_bytes(M)
    mc_bytes = point_to_bytes(message_commitment)
    kc_bytes = point_to_bytes(key_commitment)

    seed = derive_seed(sk.public_key.spki, ct.to_asn1(), message_bytes, mc_bytes, kc_bytes)
    challenge = drbg.randbelow(seed, sk.curve.q)

    response = (challenge * sk.x + t) % sk.curve.q
    return DecryptionProof(message_commitment, key_commitment, response)



def legendre(a, p):
    return pow(a, (p - 1) // 2, p)


# https://www.ctfrecipes.com/cryptography/general-knowledge/maths/modular-arithmetic/tonelli-shanks
def tonelli(a, p):
    assert legendre(a, p) == 1, "not a quadratic-residue"
    q = p - 1
    s = 0
    while q % 2 == 0:
        q //= 2
        s += 1

    if s == 1:
        return pow(a, (p + 1) // 4, p)

    for z in range(2, p):
        if p - 1 == legendre(z, p):
            break

    c = pow(z, q, p)
    r = pow(a, (q + 1) // 2, p)
    t = pow(a, q, p)

    m = s

    t2 = 0

    while (t - 1) % p != 0:
        t2 = (t * t) % p
        for i in range(1, m):
            if (t2 - 1) % p == 0:
                break
            t2 = (t2 * t2) % p
        b = pow(c, 1 << (m - i - 1), p)
        r = (r * b) % p
        c = (b * b) % p
        t = (t * c) % p
        m = i
    return r



import hashlib
import math

HASH_ALGO = hashlib.sha256


# https://github.com/valimised/ivxv/blob/published/common/java/src/main/java/ee/ivxv/common/crypto/rnd/DPRNG.java
# https://github.com/valimised/ivxv/blob/published/common/java/src/main/java/ee/ivxv/common/math/IntegerConstructor.java#L20
def randbelow(seed: bytes, n: int) -> int:
    """Return a pseudorandom int in the range [0, n)."""
    bound_size = (n.bit_length() + 7) // 8
    mask_len = n.bit_length() % 8
    if mask_len == 0:
        mask_len = 8

    block_len = HASH_ALGO().digest_size

    counter = 0
    buffer = b""

    rand = n
    while not (rand < n):
        # Compute the minimum number of hash outputs needed to meet the required byte-count.
        blocks_needed = math.ceil((bound_size - len(buffer)) / block_len)

        for i in range(blocks_needed):
            counter += 1
            buffer += HASH_ALGO(counter.to_bytes(8, "big") + seed).digest()

        # Mask off the highest byte in order to salvage some bits.
        # See also 'Bitmask with rejection (Unbiased)' in
        # https://www.pcg-random.org/posts/bounded-rands.html
        high_byte = buffer[0] & ((1 << mask_len) - 1)
        tmp_buffer = bytes([high_byte]) + buffer[1:bound_size]
        rand = int.from_bytes(tmp_buffer, "big", signed=False)

        # Clear the consumed bytes form the buffer.
        buffer = buffer[bound_size:]

    return rand



from pyivxv.crypto.keys import generate_private_key, PrivateKey, PublicKey


def test_read_private_key():
    sk = generate_private_key()
    sk_der = sk.private_bytes("DER")
    sk_pem = sk.private_bytes("PEM")

    sk1 = PrivateKey.from_private_bytes(sk_der, sk.election_id)
    sk2 = PrivateKey.from_private_bytes(sk_pem, sk.election_id)
    assert sk.x == sk1.x == sk2.x


def test_read_public_key():
    pk = generate_private_key().public_key
    pk_der = pk.public_bytes("DER")
    pk_pem = pk.public_bytes("PEM")

    pk1 = PublicKey.from_public_bytes(pk_der)
    pk2 = PublicKey.from_public_bytes(pk_pem)

    assert pk.H == pk1.H == pk2.H



from pyasn1.type.char import GeneralString
from pyasn1.type.namedtype import NamedTypes, NamedType, OptionalNamedType
from pyasn1.type.univ import ObjectIdentifier, Sequence, OctetString, Integer, Boolean
from pyasn1_modules import rfc5280

id_ivxv_ecc_elgamal = ObjectIdentifier("1.3.6.1.4.1.99999.1")


class ECCElGamalParameters(Sequence):
    componentType = NamedTypes(
        NamedType("curve", GeneralString()),
        NamedType("electionId", GeneralString()),
        # Legacy (non-lifted) ECC ElGamal keys did not have the lifted field, but used the same OID.
        OptionalNamedType("lifted", Boolean()),
    )


class ECCElGamalPublicKey(Sequence):
    componentType = NamedTypes(NamedType("pubY", OctetString()))


class ECCElGamalCiphertext(Sequence):
    componentType = NamedTypes(
        NamedType("uBlind", OctetString()),
        NamedType("vBlindedMessage", OctetString())
    )


class ElGamalCiphertextInfo(Sequence):
    componentType = NamedTypes(
        NamedType("algorithm", rfc5280.AlgorithmIdentifier()),
        NamedType("ciphertext", ECCElGamalCiphertext())
    )


class ECCElGamalDecryptionProof(Sequence):
    componentType = NamedTypes(
        NamedType("aMsgCommitment", OctetString()),
        NamedType("bKeyCommitment", OctetString()),
        NamedType("sResponse", Integer())
    )


class ECCElGamalDecryptionChallenge(Sequence):
    componentType = NamedTypes(
        NamedType("niProofDomain", GeneralString()),
        NamedType("publicKey", rfc5280.SubjectPublicKeyInfo()),
        NamedType("ciphertextInfo", ElGamalCiphertextInfo()),
        NamedType("encodedPlaintext", OctetString()),
        NamedType("aMsgCommitment", OctetString()),
        NamedType("bKeyCommitment", OctetString())
    )



from fastecdsa.point import Point
from pyasn1.codec.der import encoder as der_encoder, decoder as der_decoder
from pyasn1_modules.rfc5280 import AlgorithmIdentifier

from pyivxv.asn1.schemas import ElGamalCiphertextInfo, ECCElGamalCiphertext, id_ivxv_ecc_elgamal
from pyivxv.encoding.ec_point import point_to_bytes, point_from_bytes


class ElGamalCiphertext:
    U: Point
    V: Point

    def __init__(self, U: Point, V: Point, r: int | None = None):
        self.U = U
        self.V = V
        self._r = r

    @property
    def ephemeral_random(self) -> int:
        if self._r is None:
            raise AttributeError("Ciphertext ephemeral random not known")
        return self._r

    @ephemeral_random.setter
    def ephemeral_random(self, r: int) -> None:
        self._r = r

    def unblind(self, H: Point, *, r: int | None = None) -> Point:
        if self._r is None and r is None:
            raise ValueError("Ciphertext ephemeral random not known")
        elif r is None and self._r is not None:
            # Explicitly provided random overrides the internal random of the class.
            # Up for debate: is that a sensible choice? Should it throw on conflict?
            # Allowing to provide r seems sensible in case r should be short-lived, but the class is long-lived.
            r = self._r

        blind = H * r
        return self.V - blind

    def to_asn1(self) -> ElGamalCiphertextInfo:
        eem = ECCElGamalCiphertext()
        eem["uBlind"] = point_to_bytes(self.U)
        eem["vBlindedMessage"] = point_to_bytes(self.V)

        ai = AlgorithmIdentifier()
        ai["algorithm"] = id_ivxv_ecc_elgamal

        eb = ElGamalCiphertextInfo()
        eb["algorithm"] = ai
        eb["ciphertext"] = eem

        return eb

    def to_bytes(self) -> bytes:
        return der_encoder.encode(self.to_asn1())

    @classmethod
    def from_bytes(cls, data: bytes):
        eb, _ = der_decoder.decode(data, asn1Spec=ElGamalCiphertextInfo())
        eem = eb["ciphertext"]
        return cls(point_from_bytes(eem["uBlind"]), point_from_bytes(eem["vBlindedMessage"]))
